#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "lib.h"

//#define DebugFs		"/sys/kernel/debug"
#define DebugFs		"/debug/tracing"
#define FtraceNop	"nop"
#define FtraceFunction	"function"

static int ftrace_enable(int on)
{
	int fd, err, e;
	ssize_t ssz;
	char *p;

	fd = open(DebugFs "/tracing_enabled", O_WRONLY);
	err = fd;
	if (fd < 0)
		goto out;

	p = "0";
	if (on)
		p = "1";
	ssz = write(fd, p, 1);
	if (ssz == 1)
		err = close(fd);
	else {
		err = -1;
		e = errno;
		close(fd);
		errno = e;
	}

out:
	return err;
}

static int ftrace_do(char *func, int len)
{
	int fd, err, e;
	ssize_t ssz;

	fd = open(DebugFs "/current_tracer", O_WRONLY);
	err = fd;
	if (fd < 0)
		goto out;

	ssz = write(fd, func, len);
	if (ssz == len)
		err = close(fd);
	else {
		err = -1;
		e = errno;
		close(fd);
		errno = e;
	}

out:
	return err;
}

static int ftrace_tid(pid_t tid)
{
	int fd, err, l, e;
	ssize_t ssz;
	char a[16];

	l = snprintf(a, sizeof(a), "%d\n", tid);
	fd = open(DebugFs "/set_ftrace_pid", O_WRONLY);
	err = fd;
	if (fd < 0)
		goto out;

	err = -1;
	ssz = write(fd, a, l);
	if (ssz == l)
		err = close(fd);
	else {
		err = -1;
		e = errno;
		close(fd);
		errno = e;
	}

out:
	return err;
}

int ftrace_me(ftrace_t tgt)
{
	int err;
	pid_t pid;

	switch (tgt) {
	case FT_INIT:
		err = ftrace_do(FtraceNop, sizeof(FtraceNop) - 1);
		goto out;
	case FT_STOP:
		err = ftrace_enable(0);
		goto out;
	case FT_PROCESS:
		pid = getpid();
		break;
	case FT_THREAD:
		pid = gettid();
		break;
	default:
		abort();
	}

	err = ftrace_enable(0);
	if (!err)
		err = ftrace_tid(pid);
	if (!err)
		err = ftrace_do(FtraceFunction,
				sizeof(FtraceFunction) - 1);
	if (!err)
		err = ftrace_enable(1);

out:
	return err;
}
