/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#define _REENTRANT

#include <sys/stat.h>
#include <sys/types.h>
#include <aio.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "laiotest.h"

#define EVerify(r, e) do {				\
	err = lio_listio(LIO_WAIT, pcb, NIO, sev);	\
	assert(err == r);				\
	assert(errno == e);				\
} while (0)

int do_test(char *fname, sigevent_t *sev, int expected)
{
#define NIO 1024
	int err, i, e, nio;
	ssize_t ssz;
	struct aiocb cb[NIO];
	struct aiocb *pcb[NIO];
	char *p;
	struct stat st;

	nio = 0;
	err = posix_memalign((void *)&p, 512, BUFSIZ);
	assert(!err);
	i = 0;
	pcb[i] = cb + i;
	memset(cb, 0, sizeof(*cb));
	cb->aio_fildes = open(fname, O_RDONLY | LaioFlag);
	assert(cb->aio_fildes >= 0);
	err = fstat(cb->aio_fildes, &st);
	assert(!err);
	assert(st.st_size <= BUFSIZ);
	cb->aio_buf = p;
	cb->aio_nbytes = BUFSIZ;
	cb->aio_lio_opcode = LIO_READ;
	cb->aio_sigevent.sigev_notify = SIGEV_NONE;
	for (i++; i < NIO; i++) {
		cb[i] = cb[i - 1];
		pcb[i] = cb + i;
	}

#if 0 //defined(NonLaio) || defined(Laio)
	err = lio_listio(LIO_WAIT, pcb, NIO, sev);
	assert(err);
	assert(errno == EINVAL);
	err = laio_fin();
	assert(!err);
#endif

	struct aioinit init = {
		.aio_threads	= NIO * 2,
		.aio_num	= NIO * 2
	};
	aio_init(&init);

	err = lio_listio(LIO_WAIT, pcb, NIO, sev);
	assert(!err);
	/* for WAIT, no notify */
	//nio++;
	for (i = 0; i < NIO; i++) {
		e = aio_error(cb + i);
		assert(!e);
		ssz = aio_return(cb + i);
		assert(ssz == st.st_size);
	}

	/* suspend */
	err = lio_listio(LIO_NOWAIT, pcb, NIO, sev);
	assert(!err);
	nio++;
	for (i = 0; i < NIO; i++) {
		e = aio_error(cb + i);
		if (e == EINPROGRESS) {
			const struct aiocb *pcb;
			pcb = cb + i;
			err = aio_suspend(&pcb, 1, NULL);
			assert(!err);
		}
		ssz = aio_return(cb + i);
		assert(ssz == st.st_size);
	}

	/* cancel */
	err = lio_listio(LIO_NOWAIT, pcb, NIO, sev);
	assert(!err);
	nio++;
	for (i = 0; i < NIO; i++)
		do_cancel(cb + i, st.st_size);

	/* errors */
	close(cb[0].aio_fildes);
#if defined(NonLaio) || defined(Laio)
	EVerify(-1, EBADF);
#endif

	if (sev
	    && sev->sigev_notify != SIGEV_NONE
	    && expected)
		wait_for_all_notification(sev, expected);

	err = laio_fin();
	assert(!err);

	return nio;
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	int nio, n;
	sigevent_t sev;

	init_test();

	nio = do_test(argv[1], NULL, 0);

	set_none(&sev);
	n = do_test(argv[1], &sev, 0);
	assert(n == nio);

	set_sig(&sev);
	n = do_test(argv[1], &sev, nio);
	assert(n == nio);
	assert(n == nnotify);

	set_thread(&sev);
	n = do_test(argv[1], &sev, nio);
	assert(n == nio);
	assert(n == nnotify);

	return 0;
}
