/*
 * Copyright (C) 2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

//#define _GNU_SOURCE
#define _REENTRANT

#include <sys/time.h>
#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>

#include "lib/lib.h"
#include "laiotest.h"

#if 1
#define DPri(fmt, args...) printf("%s: " fmt, __func__, ##args)
#endif

int nnotify;
static int waiting;
static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

static void notifier(void)
{
	pthread_mutex_lock(&mtx);
	nnotify++;
	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&mtx);
}

static sigset_t sigmask;
static int signo, sigfd;
static pthread_t sigth;

static void *sig_func(void *arg)
{
	ssize_t ssz;
	struct signalfd_siginfo siginfo;

	while (1) {
		ssz = read(sigfd, &siginfo, sizeof(siginfo));
		if (ssz <= 0)
			break;
		if (ssz != sizeof(siginfo) || siginfo.ssi_signo != signo)
			continue;

#if defined(NonLaio) || defined(Laio)
		/* liblaio creates a single thread */
		if (!waiting)
			nnotify++;
		else
			notifier();
#else
		/* glibc creates multiple threads */
		notifier();
#endif
	}

	return NULL;
}

static void th_func(sigval_t sv)
{
	notifier();
}

void wait_for_all_notification(sigevent_t *sev, int expected)
{
	int err;
	struct timeval now;
	struct timespec ts;

	gettimeofday(&now, NULL);
	ts.tv_sec = now.tv_sec + 3;
	ts.tv_nsec = now.tv_usec * 1000;
	if (ts.tv_nsec > 1000 * 1000 * 1000) {
		ts.tv_nsec -= 1000 * 1000 * 1000;
		ts.tv_sec++;
	}

	err = 0;
	waiting = 1;
	pthread_mutex_lock(&mtx);
	while (nnotify < expected && err != ETIMEDOUT)
		err = pthread_cond_timedwait(&cond, &mtx, &ts);
	pthread_mutex_unlock(&mtx);
	waiting = 0;
}

void set_none(sigevent_t *sev)
{
	//DPri("enter\n");
	nnotify = 0;
	sev->sigev_notify = SIGEV_NONE;
}

void set_sig(sigevent_t *sev)
{
	int err;

	//DPri("enter\n");
	nnotify = 0;
	sigfd = signalfd(-1, &sigmask, /*flags unused*/0);
	assert(sigfd >= 0);

	err = pthread_create(&sigth, NULL, sig_func, NULL);
	assert(!err);
	err = pthread_detach(sigth);
	assert(!err);

	sev->sigev_notify = SIGEV_SIGNAL;
	sev->sigev_signo = signo;
}

void set_thread(sigevent_t *sev)
{
	//DPri("enter\n");
	nnotify = 0;
	sev->sigev_notify = SIGEV_THREAD;
	sev->sigev_notify_function = th_func;
	sev->sigev_notify_attributes = NULL;
}

void init_test(void)
{
	int err;

	signo = SIGRTMIN + 1;
	sigemptyset(&sigmask);
	sigaddset(&sigmask, signo);
	err = sigprocmask(SIG_BLOCK, &sigmask, NULL);
	assert(!err);
}

/* ---------------------------------------------------------------------- */

/* need to test more */
void do_cancel(struct aiocb *cb, ssize_t size)
{
	int err, e, e2;
	ssize_t ssz;
	const struct aiocb *pcb = cb;

	err = aio_cancel(cb->aio_fildes, cb);
	e = errno;
	if (err == AIO_CANCELED) {
		e2 = aio_error(cb);
		assert(e2 == ECANCELED);
	} else if (err == AIO_ALLDONE) {
		e2 = aio_error(cb);
		assert(!e2);
		ssz = aio_return(cb);
		assert(ssz == size);
	} else if (err == AIO_NOTCANCELED) {
		e2 = aio_suspend(&pcb, 1, NULL);
		if (!e2 || errno == EINTR) {
			e2 = aio_error(cb);
			assert(!e2);
			ssz = aio_return(cb);
			assert(ssz == size);
		} else
			assert(errno == EAGAIN);
	} else {
		printf("%d\n", err);
		assert(0);
	}

	errno = e;
}
