#!/bin/sh

# Copyright (C) 2010 J. R. Okajima

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# in KB
rd_size=$((64 * 1024))
rw_dev=/dev/ram0
ro_dev=/dev/ram1
rw=/dev/shm/rw
ro=/dev/shm/ro
mntpnt=/dev/shm/u

src="/bin /etc"
nfind=1000
# the number of files may be less
ncpup=1000

set -eu
if [ $# -ne 1 ]
then
	echo Usage: $0 'ext2|aufs|dm' 1>&2
	false
fi

alias time="time -f '%U %S %e'"

mount_ext2()
{
	sudo umount $ro
	sudo mount $ro_dev $mntpnt
}

umount_ext2()
{
	:
}

mount_aufs()
{
	local xino=/dev/shm/xino

	sudo mount -o remount,ro $ro
	sudo mkfs -t ext2 -q -F $rw_dev
	sudo mount $rw_dev $rw
	sudo mount -t aufs -o br:$rw:$ro,xino=$xino none $mntpnt
}

umount_aufs()
{
	sudo umount $rw
	sudo umount $ro
	sudo modprobe -r aufs
}

mount_dm()
{
	sudo umount $ro
	sudo modprobe dm-snapshot
	echo 0 $((${rd_size} * 2)) snapshot $ro_dev $rw_dev p 0 |
	sudo dmsetup create dm-ovl
	sudo mount /dev/mapper/dm-ovl $mntpnt
}

umount_dm()
{
	sudo dmsetup remove dm-ovl
	sudo modprobe -r dm-snapshot dm-mod
}

Mount() # type
{
	#sudo modprobe -r brd
	sudo modprobe brd rd_nr=2 rd_size=$rd_size
	sleep 1
	sudo mkfs -t ext2 -q -F $ro_dev
	sudo mount $ro_dev $ro
	sudo chmod o+w $ro
	{ tar -cf - ${src} 2> /dev/null | tar -C $ro -xpf -; } || :
	#find $ro | wc -l
	sudo chmod o-w $ro
	mkdir -p $mntpnt
	mount_$1
}

Umount() # type
{
	sudo umount $mntpnt
	umount_$1
	sudo modprobe -r brd
}

########################################
mkdir -p $rw $ro $mntpnt
Mount $1

uname -a
df $ro

# here we go
cd $mntpnt
echo find x $nfind
seq $nfind |
time xargs -i find . -xdev -name lost+found -prune -o -fprint /dev/null ||
:

echo cpup x $ncpup
find . -xdev -name lost+found -prune -o -type f |
head -n $ncpup |
time sh -c "
	while read i
	do dd if=/dev/zero of=\$i conv=notrunc oflag=append bs=1 count=1
	done > /dev/null 2>&1
	true
"

cd
Umount $1
